
;CPU Frequency

;the frequency measurement routine was written by Edgar (Donkey)
;Edgar's site: http://www.quickersoft.com/donkey/index.html

;the remainder of the program was written by DednDave

;many of the macros and libraries are provided by the MASM32 project
;thanks to Hutch and all those who contributed
;MASM32 site: http://www.masm32.com

;--------------------------------------------------------------------------

        INCLUDE    \masm32\include\masm32rt.inc
        .586

;--------------------------------------------------------------------------

EnumerateCPUs PROTO
CPUSpeed      PROTO

;--------------------------------------------------------------------------

        .DATA

Banner  db 13,10
CWTitle db 'CPU Frequency by Edgar (Donkey) and Dave (DednDave)'
BanTerm db 0,10,13,10,0
;end of Banner

AscBuf  db 21 dup(0)

        .DATA?

hStdOut dd ?              ;std output handle
hStdInp dd ?              ;std input handle
IniAttr dd ?              ;initial screen attribute
ConMode dd ?              ;initial console mode
EvntCnt dd ?              ;keyboard event count
MenuPos dd ?              ;cursor location of "Press any key to exit"
FreqPos dd ?              ;frequency display position

CursInfo   CONSOLE_CURSOR_INFO <>
BufInfo    CONSOLE_SCREEN_BUFFER_INFO <>
InpEvntRec INPUT_RECORD <>

;--------------------------------------------------------------------------

        .CODE

_main   PROC

;initialize the display

        call    SInit

;enumerate the CPUs
;this function returns the mask for the first core that supports RDTSC

        INVOKE  EnumerateCPUs
        push    eax                            ;save on stack

;display the "Frequency:" title

        print   chr$(13,10,'CPU Frequency: ')

;set the display positions

        INVOKE  GetConsoleScreenBufferInfo,
                hStdOut,
                ADDR BufInfo
        mov     eax,BufInfo.dwCursorPosition
        mov     FreqPos,eax
        push    eax
        xor     ax,ax
        add     eax,20000h
        mov     MenuPos,eax
        pop     eax

;display "MHz" and the "menu" message

        add     ax,13
        INVOKE  SetConsoleCursorPosition,
                hStdOut,
                eax
        print   chr$('MHz',13,10,13,10)

;display the "menu" message

        print   chr$('Press any key to exit')

;verify that RDTSC is supported

        pop     eax                            ;recall from stack
        or      eax,eax
        jnz     Freq01

        mov     eax,FreqPos
        xor     ax,ax
        INVOKE  SetConsoleCursorPosition,
                hStdOut,
                eax
        print   chr$('Processor does not support RDTSC')

Freq00: INVOKE  Sleep,
                100
        call    InKyb
        jnz     Exit00

        jmp     Freq00

;restrict execution to a single core

Freq01: push    eax
        INVOKE  GetCurrentProcess
        pop     edx
        INVOKE  SetProcessAffinityMask,
                eax,
                edx

;repeat display the frequency until a key is pressed

Freq02: INVOKE  CPUSpeed
        call    Asc64
        call    InKyb
        jnz     Exit00

        mov     esi,edi
        mov     edx,edi
        mov     ecx,offset AscBuf+14
        sub     ecx,edi
        jbe     Freq03

        dec     edi
        dec     edx
        rep     movsb
        mov     al,2Eh
        stosb

Freq03: push    edx
        INVOKE  SetConsoleCursorPosition,
                hStdOut,
                FreqPos
        pop     eax
        print   eax
        INVOKE  Sleep,
                1000
        jmp     Freq02

;exit the program

Exit00: call    SExit
        exit

_main   ENDP

;--------------------------------------------------------------------------

CPUSpeed PROC uses ebx edi esi

;written by Edgar (Donkey)

LOCAL hProcess :DWORD
LOCAL hThread :DWORD
LOCAL PriorityClass :DWORD
LOCAL Priority :DWORD
LOCAL timer :QWORD
LOCAL PerformanceCount :QWORD
LOCAL Frequency :QWORD

; Returns the number of clocks per second in EDX:EAX

invoke GetCurrentProcess
mov [hProcess],eax
invoke GetCurrentThread
mov [hThread],eax
invoke GetPriorityClass,[hProcess]
mov [PriorityClass],eax
invoke GetThreadPriority,[hThread]
mov [Priority],eax

invoke QueryPerformanceFrequency,addr Frequency
or eax,eax
jnz @F
invoke SetLastError,21
xor edx,edx
xor eax,eax
ret
@@:

mov esi,DWORD PTR [Frequency]
mov edi,DWORD PTR [Frequency+4]

invoke SetPriorityClass,[hProcess], REALTIME_PRIORITY_CLASS
invoke SetThreadPriority,[hThread], THREAD_PRIORITY_TIME_CRITICAL
invoke Sleep,0

invoke QueryPerformanceCounter,addr PerformanceCount
mov ebx,DWORD PTR [PerformanceCount]

@@:
invoke QueryPerformanceCounter,addr PerformanceCount
mov eax,DWORD PTR [PerformanceCount]
cmp ebx,eax
je @B
add esi,eax
adc edi,DWORD PTR [PerformanceCount+4]

ALIGN 4
rdtsc
mov DWORD PTR [timer],eax
mov DWORD PTR [timer+4],edx

@@:
invoke QueryPerformanceCounter,addr PerformanceCount
cmp edi,DWORD PTR [PerformanceCount+4]
ja @B
cmp esi,DWORD PTR [PerformanceCount]
ja @B

rdtsc
sub eax,DWORD PTR [timer]
sbb edx,DWORD PTR [timer+4]
push eax
push edx
invoke SetThreadPriority, [hThread], [Priority]
invoke SetPriorityClass, [hProcess], [PriorityClass]
invoke SetLastError,0

pop edx
pop eax

or eax,eax
RET
CPUSpeed ENDP

;--------------------------------------------------------------------------

        OPTION  PROLOGUE:NONE
        OPTION  EPILOGUE:NONE

EnumerateCPUs PROC

;Enumerate processor cores - written by DednDave
;this is a preliminary version - to be updated soon
;I was very careful to leave a lot of room for improvement - lol

;Call With: nothing (future versions will have mode controls)

;  Returns: displays a line of text describing total system cores
;           displays a line of text for each CPU package
;           eax = mask for first core that supports CPUID and RDTSC
;           DF = 0 (up)
;           ebx, ebp, esi, edi preserved

;preserve registers and initialize stack frame

;[ebp] = mask for first core that supports CPUID and RDTSC (init = 0)
;[ebp-4] = CPU number (init = 0)
;[ebp-8] = rotated affinity mask for testing (init = 1)
;[ebp-12] = current process handle
;[ebp-16] = system affinity mask
;[ebp-20] = process affinity mask/cpuid[1]eax
;[ebp-24] = total system cores
;[ebp-28] = cpuid[1]edx
;[ebp-32] = cpuid[1]ecx
;[ebp-36] = cpuid[1]ebx

        push    ebx
        push    esi
        push    edi
        push    ebp
        xor     eax,eax
        push    eax
        mov     ebp,esp
        push    eax
        inc     eax
        push    eax
        sub     esp,28

;get and store current process handle

        INVOKE  GetCurrentProcess
        mov     [ebp-12],eax

;get and store affinity masks

        lea     ecx,[ebp-20]
        lea     edx,[ebp-16]
        INVOKE  GetProcessAffinityMask,
                eax,
                ecx,
                edx

;enumerate total system cores

        print   chr$('Total System Processor Cores: ')
        mov     eax,32
        mov     edx,[ebp-16]
        mov     ecx,eax

Ecpus0: shr     edx,1
        jc      Ecpus1

        dec     eax

Ecpus1: dec     ecx
        jnz     Ecpus0

        mov     [ebp-24],eax
        print   str$(eax),13,10

;enumerate individual CPUs

Ecpus2: print   chr$('CPU ')
        print   str$(dword ptr [ebp-4]),58,32

;restrict execution to the selected core

        INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                dword ptr [ebp-8]

;verify CPUID supported

        pushfd
        pop     eax
        mov     edx,eax             ;edx=original
        xor     eax,200000h         ;eax=toggled
        push    eax
        popfd                       ;eflags=new
        pushfd
        pop     eax                 ;eax=new
        push    edx
        popfd                       ;eflags=original
        xor     eax,edx             ;eax=toggle result
        test    eax,200000h
        jnz     Ecpus3

        print   chr$('Unknown Cores: 1'),13,10
        mov     ecx,1
        jmp     EcpusR

Ecpus3: mov     eax,1
        cpuid
        mov     [ebp-20],eax
        mov     [ebp-36],ebx
        mov     [ebp-28],edx
        mov     [ebp-32],ecx
        mov     eax,80000000h
        cpuid
        cmp     eax,80000004h
        jae     Ecpus5

;if no brand string, we show fam/mod/xfam/xmod/type/step

Ecpus4: print   chr$('Fam ')
        mov     eax,[ebp-20]
        shr     eax,8
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('Mod ')
        mov     eax,[ebp-20]
        shr     eax,4
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('xFam ')
        mov     eax,[ebp-20]
        shr     eax,20
        and     eax,0FFh
        print   str$(eax),32
        print   chr$('xMod ')
        mov     eax,[ebp-20]
        shr     eax,16
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('Type ')
        mov     eax,[ebp-20]
        shr     eax,12
        and     eax,3
        print   str$(eax),32
        print   chr$('Step ')
        mov     eax,[ebp-20]
        and     eax,0Fh
        print   str$(eax),32
        jmp     Ecpus8

;brand string

Ecpus5: xor     eax,eax
        push    eax
        mov     eax,80000004h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax
        mov     eax,80000003h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax
        mov     eax,80000002h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax

;find the end

        cld
        xor     eax,eax
        mov     edi,esp
        mov     ecx,49
        repnz   scasb
        dec     edi
        dec     edi

;strip spaces off the end

        std
        mov     al,32
        lea     ecx,[edi+1]
        sub     ecx,esp
        repz    scasb
        jz      Ecpus6

        inc     edi

;terminate it with a space and a null

Ecpus6: mov     [edi+1],ax
        lea     esi,[edi+2]
        cld
        mov     edi,esp
        mov     ecx,48
        repz    scasb
        dec     edi

;strip out (r)'s, (tm)'s, and extra spaces
;(code to be added later)



;if less than 2 characters, give them the family/model string instead
;edi = offset of string
;esi = offset of terminator

        sub     esi,edi
        cmp     esi,2
        jbe     Ecpus4

Ecpus7: print   edi
        add     esp,52

;mmx, sse extensions
;(ia32/ia64/EM64T/AMD64 to be added later)
;(possibly HTT and SAHF if line length permits)

Ecpus8: xor     eax,eax
        cpuid
        mov     edx,16800010h
        and     edx,[ebp-28]
        mov     eax,80201h
        cmp     ecx,6C65746Eh   ;is it Intel ?
        jnz     Ecpus9

        or      eax,100000h

Ecpus9: and     eax,[ebp-32]
        or      edx,eax
        cmp     ecx,444D4163h   ;is it AMD ?
        jnz     EcpusA

        push    edx
        mov     eax,80000001h
        cpuid
        pop     eax
        and     ecx,840h
        and     edx,0C0400000h
        or      edx,eax
        or      edx,ecx

EcpusA: mov     ebx,edx
        shr     ebx,18
        mov     ax,dx
        and     ax,0A41h
        and     dl,10h
        or      ebx,eax
        shr     edx,1
        or      bl,dl

;at this point, 14 meaningful bits in BX:

;  Bit Source      Meaning
;   0  0_1[ecx:00] SSE3
;   1  0_1[ecx:19] SSE4.1
;   2  0_1[ecx:20] SSE4.2  (Intel only)
;   3  0_1[edx:04] RDTSC
;   4  8_1[edx:22] MMX+    (AMD only)
;   5  0_1[edx:23] MMX
;   6  8_1[ecx:06] SSE4a   (AMD only)
;   7  0_1[edx:25] SSE
;   8  0_1[edx:26] SSE2
;   9  0_1[ecx:09] SSSE3
;  10  0_1[edx:10] HTT
;  11  8_1[ecx:11] SSE5    (AMD only)
;  12  8_1[edx:30] 3DNow!+ (AMD only)
;  13  8_1[edx:31] 3DNow!  (AMD only)
;  14  0
;  15  0

;mmx extensions

        test    bl,10h
        jz      EcpusB

        print   chr$('MMX+ ')
        jmp short EcpusC

EcpusB: test    bl,20h
        jz      EcpusP          ;if no mmx, no sse or 3dnow either

        print   chr$('MMX ')

;sse extensions

EcpusC: test    bh,8
        jz      EcpusD

        print   chr$('SSE5 ')
        jmp short EcpusK

EcpusD: test    bl,40h
        jz      EcpusE

        print   chr$('SSE4a ')
        jmp short EcpusK

EcpusE: test    bl,4
        jz      EcpusF

        print   chr$('SSE4.2 ')
        jmp short EcpusP        ;if sse4.2 it's Intel, so no 3dnow

EcpusF: test    bl,2
        jz      EcpusG

        print   chr$('SSE4.1 ')
        jmp short EcpusK

EcpusG: test    bh,2
        jz      EcpusH

        print   chr$('SSSE3 ')
        jmp short EcpusK

EcpusH: test    bl,1
        jz      EcpusI

        print   chr$('SSE3 ')
        jmp short EcpusK

EcpusI: test    bh,1
        jz      EcpusJ

        print   chr$('SSE2 ')
        jmp short EcpusK

EcpusJ: test    bl,80h
        jz      EcpusK

        print   chr$('SSE ')

;3dnow extensions

EcpusK: test    bh,10h
        jz      EcpusL

        print   chr$('3DNow!+ ')
        jmp short EcpusP

EcpusL: test    bh,20h
        jz      EcpusP

        print   chr$('3DNow! ')

;count cores in package

EcpusP: mov     ecx,1
        test    bh,4
        jz      EcpusQ

        mov     ecx,[ebp-36]
        shr     ecx,16
        mov     ch,0

EcpusQ: push    ecx
        print   chr$('Cores: ')
        mov     eax,[esp]
        print   str$(eax),13,10
        pop     ecx

;next core

EcpusR: mov     eax,[ebp-8]

EcpusS: shl     eax,1
        jc      EcpusT

        mov     edx,[ebp-16]
        cmp     eax,edx
        ja      EcpusT

        test    eax,edx
        jz      EcpusS

        dec     ecx
        jnz     EcpusS

        mov     [ebp-8],eax
        inc dword ptr [ebp-4]
        jmp     Ecpus2

;find first core that supports rdtsc
;(temporary - this will be incorporated into the code above later on)

EcpusT: mov     eax,1
        mov     [ebp-8],eax

;restrict execution to the selected core

EcpusU: INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                eax

;verify CPUID supported

        pushfd
        pop     eax
        mov     edx,eax             ;edx=original
        xor     eax,200000h         ;eax=toggled
        push    eax
        popfd                       ;eflags=new
        pushfd
        pop     eax                 ;eax=new
        push    edx
        popfd                       ;eflags=original
        xor     eax,edx             ;eax=toggle result
        test    eax,200000h
        jz      EcpusW

;test for rdtsc

        mov     eax,1
        cpuid
        test    dl,10h
        jz      EcpusW

;test the counter- give it 2 chances to pass

        mov     ecx,2

EcpusV: push    ecx
        xor     eax,eax
        cpuid
        rdtsc
        push    eax
        xor     eax,eax
        cpuid
        rdtsc
        pop     edx
        pop     ecx
        cmp     eax,edx
        ja      EcpusY

        dec     ecx
        jnz     EcpusV

;next core

EcpusW: mov     eax,[ebp-8]

EcpusX: shl     eax,1
        jc      EcpusZ

        mov     edx,[ebp-16]
        cmp     eax,edx
        ja      EcpusZ

        test    eax,edx
        jz      EcpusX

        mov     [ebp-8],eax
        jmp     EcpusU

;mark it as usable and exit

EcpusY: mov     eax,[ebp-8]
        mov     [ebp],eax

;exit the routine

EcpusZ: INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                dword ptr [ebp-16]
        mov     esp,ebp
        pop     eax
        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
        ret

EnumerateCPUs ENDP

        OPTION  PROLOGUE:PrologueDef
        OPTION  EPILOGUE:EpilogueDef

;--------------------------------------------------------------------------

Asc64   PROC

;Asc64 by DednDave
;Convert 64-bit unsigned integer to ASCII decimal string
;
;this represents the "standard" method with only the AAM enhancement added
;this is my current version - to be improved upon
;it uses multiple-precision division by 100, then uses AAM to convert 2 digits at once
;converting 2 digits per loop pass is faster, of course (half as many loop passes)
;plus it yields a hidden advantage when it comes time to suppress a leading zero
;i.e. if the result is "0", no need to test for anti-suppression
;
;Call With: EDX:EAX= QWORD value to convert
;
;  Returns: EDI= offset into AscBuf of first numchar

        std
        mov     edi,offset AscBuf+18
        mov     ecx,edx
        xchg    eax,esi
        mov     ebx,100

Asc64a: xor     edx,edx
        xchg    eax,ecx
        div     ebx
        xchg    eax,ecx
        xchg    eax,esi
        div     ebx
        xchg    eax,esi
        xchg    eax,edx
        aam
        xchg    al,ah
        or      ax,3030h
        stosw
        mov     eax,ecx
        or      eax,esi
        jnz     Asc64a

        inc     edi
        inc     edi
        cld
        cmp byte ptr [edi],30h       ;leading 0 ?
        jz      Asc64b               ;yes - supress it

        ret                          ;no - done

Asc64b: inc     edi
        ret

Asc64   ENDP

;--------------------------------------------------------------------------

InKyb   PROC

;Poll keyboard

;  Returns: if no key: eax = 0
;                      ZF = set (zero)
;
;              if key: eax = virtual scan code
;                      ZF = clear (not zero)
;
InKyb0: push    eax
        INVOKE  PeekConsoleInput,
                hStdInp,
                ADDR InpEvntRec,
                1,
                esp
        pop     eax
        or      eax,eax
        jz      InKyb1             ;ZF set if no key

        INVOKE  ReadConsoleInput,
                hStdInp,
                ADDR InpEvntRec,
                1,
                ADDR EvntCnt
        dec word ptr InpEvntRec.EventType
        jnz     InKyb0             ;non-keyboard event type - check again

        dec dword ptr InpEvntRec.KeyEvent.bKeyDown
        jnz     InKyb0             ;keypress only

        xor     eax,eax
        inc     eax                ;ZF cleared if key
        movzx   eax,word ptr InpEvntRec.KeyEvent.wVirtualScanCode

InKyb1: ret

InKyb   ENDP

;--------------------------------------------------------------------------

SInit   PROC

;set console window title

        INVOKE  SetConsoleTitle,
                ADDR CWTitle
        mov byte ptr BanTerm,13

;save initial cursor visibility value
;set cursor visibility to false (off)

        INVOKE  GetStdHandle,
                STD_OUTPUT_HANDLE
        mov     hStdOut,eax
        INVOKE  GetConsoleCursorInfo,
                eax,
                ADDR CursInfo
        push    CursInfo.bVisible
        mov dword ptr CursInfo.bVisible,FALSE
        INVOKE  SetConsoleCursorInfo,
                hStdOut,
                ADDR CursInfo
        pop     CursInfo.bVisible

;save initial screen attribute

        INVOKE  GetConsoleScreenBufferInfo,
                hStdOut,
                ADDR BufInfo
        movzx   eax,BufInfo.wAttributes
        mov     IniAttr,eax

;set text foregound to high intensity green

        INVOKE  SetConsoleTextAttribute,
                hStdOut,
                FOREGROUND_GREEN or FOREGROUND_INTENSITY

;display the banner

        mov     eax,offset Banner
        print   eax

;get/store the standard input handle

        INVOKE  GetStdHandle,
                STD_INPUT_HANDLE
        mov     hStdInp,eax

;save the initial console mode

        INVOKE  GetConsoleMode,
                eax,
                ADDR ConMode

;disable most input modes for this type of display

SInitB: INVOKE  SetConsoleMode,
                hStdInp,
                ENABLE_EXTENDED_FLAGS

;flush input buffer

        INVOKE  FlushConsoleInputBuffer,
                hStdInp
        ret

SInit   ENDP

;--------------------------------------------------------------------------

SExit   PROC

;restore screen attribute to initial setting

        INVOKE  SetConsoleTextAttribute,
                hStdOut,
                IniAttr

;overwrite the "menu" message

        INVOKE  SetConsoleCursorPosition,
                hStdOut,
                MenuPos
        print   chr$('                     ')

;restore initial console mode

        INVOKE  SetConsoleMode,
                hStdInp,
                ConMode

;flush input buffer

        INVOKE  FlushConsoleInputBuffer,
                hStdInp

;position cursor for exit

        mov     eax,MenuPos
        sub     eax,10000h
        INVOKE  SetConsoleCursorPosition,
                hStdOut,
                eax

;set cursor visibility to initial setting

        INVOKE  SetConsoleCursorInfo,
                hStdOut,
                ADDR CursInfo
        ret

SExit   ENDP

;--------------------------------------------------------------------------

        END     _main
